/*
 * EffectsControl.c
 *
 * Created: 14.12.2016 18:43:54
 *  Author: Thomas
 */ 

#include <avr/io.h>
#include <stddef.h>
#include "../BasicDefines.h"

#include "EffectsControl.h"
#include "CandelLightEffect.h"
#include "WaveEffect.h"
#include "RainbowEffect.h"
#include "SnowEffect.h"
#include "../main.h"
#include "../HSV2RGB/HSV2RGB.h"
#include "../APA102/APA102.h"
#include "../GammaCorrection/GammaCorrection.h"


extern uint8_t ClockPulse10ms;
extern uint8_t ClockPulse1s;

extern volatile uint8_t UpdateLEDsIsAktive;
extern uint8_t SelectedMode;

extern CandelLightEffect_T CandelLightEffectStripe[LED_COUNT];
extern WaveEffect_T WaveEffectStripe[LED_COUNT];
extern RainbowEffect_T RainbowEffectStripe[LED_COUNT];
extern SnowEffect_T SnowEffectStripe[LED_COUNT];

extern APA102Frame_T APA102Frame;

extern uint8_t PotiHSV_H;
extern uint8_t PotiHSV_S;
extern uint8_t PotiBrightness;


uint8_t DebugLEDs;
uint8_t AutomaticColorChangePulse;


typedef enum tagEffects {
	CandelLight,
	Snow,
	Wave,
	Rainbow,
	All,
	None,
} Effects_T, *pEffects_T;

typedef struct tagStripeBGR {
	BGR_T LED[LED_COUNT+LED_COUNT_1FLOOR];
} StripeBGR_T, *pStripeBGR_T;

typedef enum tagEffectsControlStateMachine {
	WaitForTrigger,
	WaitStreamFinished,
} EffectsControlStateMachine_T, *pEffectsControlStateMachine_T;


typedef void(*EffectFunction)(uint8_t colorShift);
typedef void(*EffectHSV2BGRFunction)(pStripeBGR_T stripe);


static void DisplayAutomaticColorChangeStatus(void);
static void FadingPrevious2CurrentEffect(void);
static void EmptyFunctionDummy(void);
static EffectFunction GetEffectFunctionFromEffect(Effects_T effect);
static EffectHSV2BGRFunction GetEffectHSV2BGRFunctionFromEffect(Effects_T effect);


static StripeBGR_T CurrentStripeBGR;
static StripeBGR_T PreviousStripeBGR;

static Effects_T CurrentEffect;
static Effects_T CurrentEffectAll;
static Effects_T LastCurrentEffect;
static Effects_T LastEffect;
static uint16_t FadingEffectChanged;
static uint8_t EffectChangedManual;

static uint8_t AutomaticColorChange;
static uint8_t CurrentAutomaticColorShift;
static uint8_t PreviousAutomaticColorShift;

static EffectsControlStateMachine_T EffectsControltateMachine;

static EffectFunction CurrentEffectFunction;
static EffectFunction PreviousEffectFunction;
static EffectHSV2BGRFunction CurrentEffectHSV2BGRFunction;
static EffectHSV2BGRFunction PreviousEffectHSV2BGRFunction;


void EffectsControlInit()
{
	APA102Init();
	GammaCorrectionInit();
	CandelLightEffectInit();
	WaveEffectInit();
	RainbowEffectInit();
	SnowEffectInit();
	
	for (uint8_t i = 0; i < LED_COUNT; i++)
	{
		APA102Frame.LED[i].Global = BRIGHTNESS_TO_GLOBAL(BRIGHTNESS_GLOBAL);
	}
	
	CurrentEffectFunction = GetEffectFunctionFromEffect(CandelLight);
	CurrentEffectHSV2BGRFunction = GetEffectHSV2BGRFunctionFromEffect(CandelLight);

	PreviousEffectFunction = GetEffectFunctionFromEffect(None);
	PreviousEffectHSV2BGRFunction = GetEffectHSV2BGRFunctionFromEffect(None);

	EffectsControltateMachine = WaitStreamFinished;
}


void EffectsControl()
{
	static uint8_t DelayChangeEffect;
	
	uint8_t tmpCurrentEffect = CurrentEffect;
	
	if (tmpCurrentEffect != LastCurrentEffect)
	{
		LastCurrentEffect = tmpCurrentEffect;
		EffectChangedManual = TRUE;
	}
	
	
	if (tmpCurrentEffect == All)
	{
		if (ClockPulse1s)
		{
			if (++DelayChangeEffect == AUTOMATIC_INTERVAL_TIME)
			{
				DelayChangeEffect = 0;
				
				#ifndef NO_RAINBOW_AUTOMATIC
					if (++CurrentEffectAll == All)
				#else
					if (++CurrentEffectAll == Rainbow)
				#endif
				{
					CurrentEffectAll = (Effects_T)0;
				}
			}
		}
		
		tmpCurrentEffect = CurrentEffectAll;
	}
	
	
	if (tmpCurrentEffect != LastEffect)
	{
		CurrentEffectFunction = GetEffectFunctionFromEffect(tmpCurrentEffect);
		CurrentEffectHSV2BGRFunction = GetEffectHSV2BGRFunctionFromEffect(tmpCurrentEffect);

		PreviousEffectFunction = GetEffectFunctionFromEffect(LastEffect);
		PreviousEffectHSV2BGRFunction = GetEffectHSV2BGRFunctionFromEffect(LastEffect);
		
		LastEffect = tmpCurrentEffect;
		FadingEffectChanged = 0;

		switch (tmpCurrentEffect)
		{
			case CandelLight:
			case Wave:
			case Snow:
				CurrentAutomaticColorShift += 32;
			break;

			default:
			break;
		}
	}
		
	
	switch (EffectsControltateMachine)
	{
		case WaitForTrigger:
		{
			if (ClockPulse10ms)
			{
DEBUG_TOGGLE_PIN1;
				APA102UpdateLEDs();
				
				uint8_t tmpCurrentAutomaticColorShift = CurrentAutomaticColorShift;
				uint8_t tmpPreviousAutomaticColorShift = PreviousAutomaticColorShift;
				
				if ((CurrentEffect != All) || !AutomaticColorChange)
				{
					tmpCurrentAutomaticColorShift = 0;
					tmpPreviousAutomaticColorShift = 0;
				}
				
				CurrentEffectFunction(tmpCurrentAutomaticColorShift);
				PreviousEffectFunction(tmpPreviousAutomaticColorShift);
						
				EffectsControltateMachine = WaitStreamFinished;
DEBUG_TOGGLE_PIN1;
			}
		}
		break;
				
		case WaitStreamFinished:
		{
			if (UpdateLEDsIsAktive)
			{
				break;
			}
DEBUG_TOGGLE_PIN1;

			CurrentEffect = SelectedMode;

			CurrentEffectHSV2BGRFunction(&CurrentStripeBGR);
DEBUG_TOGGLE_PIN1;
			PreviousEffectHSV2BGRFunction(&PreviousStripeBGR);
DEBUG_TOGGLE_PIN1;

			FadingPrevious2CurrentEffect();

			DisplayAutomaticColorChangeStatus();
			GammaCorrectionBGR(CurrentStripeBGR.LED);

			for (uint8_t i = 0; i < LED_COUNT; i++)
			{
				APA102Frame.LED[i].Blue = CurrentStripeBGR.LED[i].Blue;
				APA102Frame.LED[i].Green = CurrentStripeBGR.LED[i].Green;
				APA102Frame.LED[i].Red = CurrentStripeBGR.LED[i].Red;
			}
				
			// Copy neighbor LEDs to first Floor LEDs
			for (uint8_t i = 0; i < (LED_COUNT_1FLOOR>>1); i++)
			{
				APA102Frame.LED[LED_COUNT+(LED_COUNT_1FLOOR>>1)+i] = APA102Frame.LED[LED_START_1FLOOR+i];
				APA102Frame.LED[LED_COUNT+i] = APA102Frame.LED[LED_COUNT-(LED_COUNT_1FLOOR>>1)-LED_START_1FLOOR+i];
			}
			
			EffectsControltateMachine = WaitForTrigger;


			// Show 8 Bit debug infos with first 8 LEDs
			#ifdef DEBUG_LEDS
				//DebugLEDs = PotiBrightness;

				for (uint8_t i = 0; i < 8; i++)
				{
					APA102Frame.LED[i].Blue = 0;
					APA102Frame.LED[i].Green = 0;
					APA102Frame.LED[i].Red = (DebugLEDs & (1<<i)) ? 64 : 0;
				}
			#endif
			
DEBUG_TOGGLE_PIN1;
		}
		break;
	}
}


static void FadingPrevious2CurrentEffect(void)
{
	if (FadingEffectChanged >= 256)
	{
		return;
	}
	

	if (EffectChangedManual)
	{
		FadingEffectChanged += 4;
	}
	else
	{
		FadingEffectChanged += 1;
	}
				
				
	if (FadingEffectChanged >= 256)
	{
		PreviousEffectFunction = GetEffectFunctionFromEffect(None);
		PreviousEffectHSV2BGRFunction = GetEffectHSV2BGRFunctionFromEffect(None);
					
		EffectChangedManual = FALSE;
					
		PreviousAutomaticColorShift = CurrentAutomaticColorShift;
	}
	else
	{
		uint8_t tmpFadingEffectChanged = FadingEffectChanged;
		uint8_t tmpPreviousFading = 256 - tmpFadingEffectChanged;

		for (uint8_t i = 0; i < LED_COUNT; i++)
		{
			CurrentStripeBGR.LED[i].Blue = (((uint16_t)CurrentStripeBGR.LED[i].Blue * tmpFadingEffectChanged) >> 8)
				+ (((uint16_t)PreviousStripeBGR.LED[i].Blue * tmpPreviousFading) >> 8);

			CurrentStripeBGR.LED[i].Green = (((uint16_t)CurrentStripeBGR.LED[i].Green * tmpFadingEffectChanged) >> 8)
				+ (((uint16_t)PreviousStripeBGR.LED[i].Green * tmpPreviousFading) >> 8);

			CurrentStripeBGR.LED[i].Red = (((uint16_t)CurrentStripeBGR.LED[i].Red * tmpFadingEffectChanged) >> 8)
				+ (((uint16_t)PreviousStripeBGR.LED[i].Red * tmpPreviousFading) >> 8);
		}
	}
	
}


static void EmptyFunctionDummy(void)
{
}


static EffectFunction GetEffectFunctionFromEffect(Effects_T effect)
{
	switch (effect)
	{
		case CandelLight:
			return &CandelLightEffect;
		break;
			
		case Wave:
			return &WaveEffect;
		break;

		case Rainbow:
			return (EffectFunction)&RainbowEffect;
		break;
			
		case Snow:
			return &SnowEffect;
		break;
			
		default:
			return (EffectFunction)&EmptyFunctionDummy;
		break;
	}
}


static void HSV2BGRCandelLightEffect(pStripeBGR_T stripe)
{
	for (uint8_t i = 0; i < LED_COUNT; i++)
	{
		HSV2BGR(&CandelLightEffectStripe[i].HSV, &stripe->LED[i]);
	}
}

static void HSV2BGRWaveEffect(pStripeBGR_T stripe)
{
	for (uint8_t i = 0; i < LED_COUNT; i++)
	{
		HSV2BGR(&WaveEffectStripe[i].HSV, &stripe->LED[i]);
	}
}

static void HSV2BGRRainbowEffect(pStripeBGR_T stripe)
{
	for (uint8_t i = 0; i < LED_COUNT; i++)
	{
		HSV2BGR(&RainbowEffectStripe[i].HSV, &stripe->LED[i]);
	}
}

static void HSV2BGRSnowEffect(pStripeBGR_T stripe)
{
	for (uint8_t i = 0; i < LED_COUNT; i++)
	{
		HSV2BGR(&SnowEffectStripe[i].HSV, &stripe->LED[i]);
	}
}

static EffectHSV2BGRFunction GetEffectHSV2BGRFunctionFromEffect(Effects_T effect)
{
	switch (effect)
	{
		case CandelLight:
			return &HSV2BGRCandelLightEffect;
		break;
		
		case Wave:
			return &HSV2BGRWaveEffect;
		break;

		case Rainbow:
			return &HSV2BGRRainbowEffect;
		break;
		
		case Snow:
			return &HSV2BGRSnowEffect;
		break;
		
		default:
			return (EffectHSV2BGRFunction)&EmptyFunctionDummy;
		break;
	}
}


// AutomaticColorChangeStatus handling
static void DisplayAutomaticColorChangeStatus(void)
{
	static uint8_t DurationDisplayAutomaticColorChange;
	static Effects_T LastRunnigEffect;
	
	if (AutomaticColorChangePulse)
	{
		AutomaticColorChangePulse = 0;
		
		if (CurrentEffect == All)
		{
			AutomaticColorChange ^= 0xff;
			
			DurationDisplayAutomaticColorChange = 100;
		}
	}
	
	
	if (LastRunnigEffect != CurrentEffect)
	{
		LastRunnigEffect = CurrentEffect;
		
		if (CurrentEffect == All)
		{
			DurationDisplayAutomaticColorChange = 100;
		}
	}
		
	
	if (DurationDisplayAutomaticColorChange)
	{
		DurationDisplayAutomaticColorChange--;
		
		BGR_T tmpBGRColorOff = { 0, 0, 0 };
		BGR_T tmpBGRColor = { 0, 0, 0 };
		
		if (AutomaticColorChange)
		{
			tmpBGRColor.Green = 0xff;
		}
		else
		{
			tmpBGRColor.Red = 0xff;
		}
		
		#ifndef TEST_BOARD
			CurrentStripeBGR.LED[HALF_LED_COUNT-6] = tmpBGRColorOff;
			CurrentStripeBGR.LED[HALF_LED_COUNT-5] = tmpBGRColorOff;
			CurrentStripeBGR.LED[HALF_LED_COUNT-4] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT-3] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT-2] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT-1] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT-0] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT+1] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT+2] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT+3] = tmpBGRColor;
			CurrentStripeBGR.LED[HALF_LED_COUNT+4] = tmpBGRColorOff;
			CurrentStripeBGR.LED[HALF_LED_COUNT+5] = tmpBGRColorOff;
		#else
			CurrentStripeBGR.LED[ 0] = tmpBGRColorOff;
			CurrentStripeBGR.LED[ 1] = tmpBGRColorOff;
			CurrentStripeBGR.LED[ 2] = tmpBGRColor;
			CurrentStripeBGR.LED[ 3] = tmpBGRColor;
			CurrentStripeBGR.LED[ 4] = tmpBGRColor;
			CurrentStripeBGR.LED[ 5] = tmpBGRColor;
			CurrentStripeBGR.LED[ 6] = tmpBGRColor;
			CurrentStripeBGR.LED[ 7] = tmpBGRColor;
			CurrentStripeBGR.LED[ 8] = tmpBGRColor;
			CurrentStripeBGR.LED[ 9] = tmpBGRColor;
			CurrentStripeBGR.LED[10] = tmpBGRColorOff;
			CurrentStripeBGR.LED[11] = tmpBGRColorOff;
		#endif
	}
	
}

